@@ -1,12 +1,8 @@
-use crate::{aabb::Bounds, builders::BuildAlgorithm};
-use crate::{builders::spatial_sah::SpatialTriangle, Ray};
-use crate::{builders::*, BvhIterator, MbvhIterator};
-use crate::{bvh_node::*, BvhPacketIterator};
-use crate::{mbvh_node::*, MbvhPacketIterator};
-use crate::{Aabb, RayPacket4};
+use crate::*;
 use glam::*;
-use serde::{Deserialize, Serialize};
-use std::fmt::Debug;
+use rayon::prelude::*;
+use std::fmt::{Debug, Formatter};
+use std::num::NonZeroUsize;
 
 pub trait Primitive<AabbType: Debug + Copy + Send + Sync = i32>: Debug + Send + Sync {
     fn center(&self) -> Vec3;
@@ -14,7 +10,8 @@ pub trait Primitive<AabbType: Debug + Copy + Send + Sync = i32>: Debug + Send +
     fn aabb(&self) -> Aabb<AabbType>;
 }
 
-#[derive(Debug, Copy, Clone, Serialize, Deserialize)]
+#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq, Hash)]
 pub enum BuildType {
     None,
     LocallyOrderedClustered,
@@ -22,32 +19,124 @@ pub enum BuildType {
     Spatial,
 }
 
+#[derive(Debug, Copy, Clone, Ord, PartialOrd, Eq, PartialEq, Hash)]
+pub enum BuildError {
+    NoPrimitives,
+    InequalAabbsAndPrimitives(usize, usize),
+}
+
+impl std::fmt::Display for BuildError {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        write!(
+            f,
+            "{}",
+            match self {
+                BuildError::NoPrimitives => "No primitives supplied".to_owned(),
+                BuildError::InequalAabbsAndPrimitives(bbs, prims) => {
+                    format!("#Aabbs({}) != #Primitives({})", *bbs, *prims)
+                }
+            }
+        )
+    }
+}
+
+impl std::error::Error for BuildError {}
+
+#[derive(Debug, Copy, Clone)]
 pub struct Builder<'a, T: Primitive<i32>> {
-    pub aabbs: &'a [Aabb<i32>],
+    /// Reference to aabbs
+    pub aabbs: Option<&'a [Aabb<i32>]>,
+    /// Reference to primitives
     pub primitives: &'a [T],
-    pub primitives_per_leaf: usize,
+    /// Max number of primitives that can be stored in a leaf node
+    pub primitives_per_leaf: Option<NonZeroUsize>,
 }
 
 impl<'a, T: Primitive<i32>> Builder<'a, T> {
-    pub fn construct_spatial_sah(self) -> Bvh
+    pub fn construct_spatial_sah(self) -> Result<Bvh, BuildError>
     where
         T: SpatialTriangle,
     {
-        spatial_sah::SpatialSahBuilder::new(self.aabbs, self.primitives, self.primitives_per_leaf)
-            .build()
+        if self.primitives.is_empty() {
+            return Err(BuildError::NoPrimitives);
+        }
+
+        if let Some(aabbs) = self.aabbs {
+            if aabbs.len() != self.primitives.len() {
+                return Err(BuildError::InequalAabbsAndPrimitives(
+                    aabbs.len(),
+                    self.primitives.len(),
+                ));
+            }
+
+            Ok(SpatialSahBuilder::new(aabbs, self.primitives, self.primitives_per_leaf).build())
+        } else {
+            let aabbs = self
+                .primitives
+                .iter()
+                .par_bridge()
+                .map(|p| p.aabb())
+                .collect::<Vec<_>>();
+            Ok(SpatialSahBuilder::new(&aabbs, self.primitives, self.primitives_per_leaf).build())
+        }
     }
 
-    pub fn construct_binned_sah(self) -> Bvh {
-        binned_sah::BinnedSahBuilder::new(self.aabbs, self.primitives).build()
+    pub fn construct_binned_sah(self) -> Result<Bvh, BuildError> {
+        if self.primitives.is_empty() {
+            return Err(BuildError::NoPrimitives);
+        }
+
+        if let Some(aabbs) = self.aabbs {
+            if aabbs.len() != self.primitives.len() {
+                return Err(BuildError::InequalAabbsAndPrimitives(
+                    aabbs.len(),
+                    self.primitives.len(),
+                ));
+            }
+
+            Ok(BinnedSahBuilder::new(&aabbs, self.primitives, self.primitives_per_leaf).build())
+        } else {
+            let aabbs = self
+                .primitives
+                .iter()
+                .par_bridge()
+                .map(|p| p.aabb())
+                .collect::<Vec<_>>();
+
+            Ok(BinnedSahBuilder::new(&aabbs, self.primitives, self.primitives_per_leaf).build())
+        }
     }
 
-    pub fn construct_locally_ordered_clustered(self) -> Bvh {
-        locb::LocallyOrderedClusteringBuilder::new(self.aabbs, self.primitives).build()
+    pub fn construct_locally_ordered_clustered(self) -> Result<Bvh, BuildError> {
+        if self.primitives.is_empty() {
+            return Err(BuildError::NoPrimitives);
+        }
+
+        if let Some(aabbs) = self.aabbs {
+            if aabbs.len() != self.primitives.len() {
+                return Err(BuildError::InequalAabbsAndPrimitives(
+                    aabbs.len(),
+                    self.primitives.len(),
+                ));
+            }
+
+            Ok(LocallyOrderedClusteringBuilder::new(aabbs, self.primitives).build())
+        } else {
+            let aabbs = self
+                .primitives
+                .iter()
+                .par_bridge()
+                .map(|p| p.aabb())
+                .collect::<Vec<_>>();
+
+            Ok(LocallyOrderedClusteringBuilder::new(&aabbs, self.primitives).build())
+        }
     }
 }
 
 // A BVH structure with nodes and primitive indices
-#[derive(Debug, Clone, Serialize, Deserialize)]
+#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[derive(Debug, Clone)]
 pub struct Bvh {
     pub(crate) nodes: Vec<BvhNode>,
     pub(crate) prim_indices: Vec<u32>,
@@ -163,24 +252,41 @@ impl Bvh {
         (self.nodes, self.prim_indices)
     }
 
-    pub fn traverse_iter<'a, T: Primitive>(
+    pub fn traverse_iter<'a, 'b, T: 'a + Primitive>(
         &'a self,
-        ray: &'a mut Ray,
+        ray: &'b mut Ray,
         primitives: &'a [T],
-    ) -> BvhIterator<'a, T> {
+    ) -> BvhIterator<'a, 'b, T> {
         BvhIterator::new(ray, self, primitives)
     }
 
-    pub fn traverse_iter_packet<'a, T: Primitive>(
+    pub fn traverse_iter_packet<'a, 'b, T: 'a + Primitive>(
         &'a self,
-        ray: &'a mut RayPacket4,
+        ray: &'b mut RayPacket4,
         primitives: &'a [T],
-    ) -> BvhPacketIterator<'a, T> {
+    ) -> BvhPacketIterator<'a, 'b, T> {
         BvhPacketIterator::new(ray, self, primitives)
     }
 }
 
-#[derive(Debug, Clone, Serialize, Deserialize)]
+impl<'a, 'b, T: 'a + Primitive> IntoRayIterator<'a, 'b, T> for Bvh {
+    type RIterator = BvhIterator<'a, 'b, T>;
+
+    fn iter(&'a self, ray: &'b mut Ray, primitives: &'a [T]) -> Self::RIterator {
+        BvhIterator::new(ray, self, primitives)
+    }
+}
+
+impl<'a, 'b, T: 'a + Primitive> IntoPacketIterator<'a, 'b, T> for Bvh {
+    type RIterator = BvhPacketIterator<'a, 'b, T>;
+
+    fn iter(&'a self, packet: &'b mut RayPacket4, primitives: &'a [T]) -> Self::RIterator {
+        BvhPacketIterator::new(packet, self, primitives)
+    }
+}
+
+#[cfg_attr(feature = "serde", derive(Serialize, Deserialize))]
+#[derive(Debug, Clone)]
 pub struct Mbvh {
     pub(crate) nodes: Vec<BvhNode>,
     pub(crate) m_nodes: Vec<MbvhNode>,
@@ -288,19 +394,19 @@ impl Mbvh {
         (self.m_nodes, self.prim_indices)
     }
 
-    pub fn traverse_iter<'a, T: Primitive>(
+    pub fn traverse_iter<'a, 'b, T: Primitive>(
         &'a self,
-        ray: &'a mut Ray,
+        ray: &'b mut Ray,
         primitives: &'a [T],
-    ) -> MbvhIterator<'a, T> {
+    ) -> MbvhIterator<'a, 'b, T> {
         MbvhIterator::new(ray, self, primitives)
     }
 
-    pub fn traverse_iter_packet<'a, T: Primitive>(
+    pub fn traverse_iter_packet<'a, 'b, T: Primitive>(
         &'a self,
-        ray: &'a mut RayPacket4,
+        ray: &'b mut RayPacket4,
         primitives: &'a [T],
-    ) -> MbvhPacketIterator<'a, T> {
+    ) -> MbvhPacketIterator<'a, 'b, T> {
         MbvhPacketIterator::new(ray, self, primitives)
     }
 }
@@ -322,3 +428,19 @@ impl Bounds<i32> for Mbvh {
         self.nodes[0].bounds
     }
 }
+
+impl<'a, 'b, T: 'a + Primitive> IntoRayIterator<'a, 'b, T> for Mbvh {
+    type RIterator = MbvhIterator<'a, 'b, T>;
+
+    fn iter(&'a self, ray: &'b mut Ray, primitives: &'a [T]) -> Self::RIterator {
+        MbvhIterator::new(ray, self, primitives)
+    }
+}
+
+impl<'a, 'b, T: 'a + Primitive> IntoPacketIterator<'a, 'b, T> for Mbvh {
+    type RIterator = MbvhPacketIterator<'a, 'b, T>;
+
+    fn iter(&'a self, packet: &'b mut RayPacket4, primitives: &'a [T]) -> Self::RIterator {
+        MbvhPacketIterator::new(packet, self, primitives)
+    }
+}
